import os
import requests
from openai import OpenAI
import google.generativeai as genai
import json
from termcolor import colored

class UnifiedChatBot:
    def __init__(self, max_memory=2500): # max memory is the maximum number of words in the memory
        self.memories = {
            'openai': [],
            'together': [],
            'openrouter': [],
            'perplexity': [],
            'gemini': [],
            'mistral': []
        }
        self.max_memory = max_memory
        self.clients = {
            'openai': OpenAI(api_key=self._get_api_key("OPENAI_API_KEY")),
            'together': OpenAI(api_key=self._get_api_key("TOGETHER_API_KEY"), base_url="https://api.together.xyz/v1"),
            'openrouter': OpenAI(api_key=self._get_api_key("OR_API_KEY"), base_url="https://openrouter.ai/api/v1"),
            'perplexity': {"api_key": self._get_api_key('PERPLEXITY_API_KEY'), "url": "https://api.perplexity.ai/chat/completions"},
            'gemini': None,
            'mistral': {
                "api_key": self._get_api_key('MISTRAL_API_KEY'),
                "url": "https://api.mistral.ai/v1/chat/completions"
            }
        }

        gemini_api_key = self._get_api_key('GEMINI_API_KEY')
        if gemini_api_key is not None:
            genai.configure(api_key=gemini_api_key)
            self.clients['gemini'] = genai.GenerativeModel('gemini-pro')
            # we need to initialize gemini chat here globally
            self.gemini_chat = self.clients['gemini'].start_chat(history=[])
        else:
            print(colored("Gemini API key not found. Skipping this API.", 'red'))
            self.gemini_chat = None
            self.clients['gemini'] = None

    def _get_api_key(self, key_name):
        key = os.getenv(key_name)
        if key is None or key == "YOUR KEY HERE":
            print(colored(f"API key for {key_name} not found. Please set it in environment variables or don't use this API.", 'red'))
            return None
        return key
    def add_message(self, role, content, api_type):
        self.memories[api_type].append({"role": role, "content": content})

    def trim_memory(self, api_type):
        if api_type == 'gemini':
            return
        while sum(len(m['content'].split()) for m in self.memories[api_type] if m['role'] != 'system') > self.max_memory:
            for message in self.memories[api_type]:
                if message['role'] != 'system':
                    self.memories[api_type].remove(message)

    def chat_main(self, user_input, api_type):
        self.add_message("user", user_input, api_type)
        if api_type == 'openai':
            response = self._openai_chat()
        elif api_type == 'together':
            response = self._together_chat()
        elif api_type == 'openrouter':
            response = self._openrouter_chat()
        elif api_type == 'perplexity':
            response = self._perplexity_chat()
        elif api_type == 'gemini':
            response = self._gemini_chat()
        elif api_type == 'mistral':
            response = self._mistral_chat()
        else:
            raise ValueError(f"Unsupported API type: {api_type}")
        self.add_message("assistant", response, api_type)
        self.trim_memory(api_type)
        # clear memory of api is perplexity beucase it is recommended to have one turn conversations
        if api_type == 'perplexity':
            self.memories[api_type] = []
        return response

    def _openai_chat(self):
        try:
            response = self.clients['openai'].chat.completions.create(
                model="gpt-4-1106-preview",
                messages=self.memories['openai'],
            )
        except Exception as e:
            print(f"OpenAI API Error: {e}")
        return response.choices[0].message.content
    
    def _mistral_chat(self):
        headers = {
            "accept": "application/json",
            "content-type": "application/json",
            "Authorization": f"Bearer {self.clients['mistral']['api_key']}"
        }
        payload = {
            "model": "mistral-medium",
            "messages": self.memories['mistral']
        }
        try:
            response = requests.post(self.clients['mistral']['url'], json=payload, headers=headers)
        except Exception as e:
            print(f"Mistral API Error: {e}")
        response = json.loads(response.text)
        return response["choices"][0]["message"]["content"]

    def _together_chat(self):
        # This ones gives an error which I couldn't figure out. the standalone version works fine in file together_mixtral.py
        try:
            response = self.clients['together'].chat.completions.create(
                model="mistralai/Mixtral-8x7B-Instruct-v0.1",
                messages=self.memories['together'],
                # max_tokens=1024,
            )
        except Exception as e:
            print(f"Together API Error: {e}")
        return response.choices[0].message.content

    def _openrouter_chat(self):
        try:
            completion = self.clients['openrouter'].chat.completions.create(
                model="mistralai/mixtral-8x7b-instruct",
                messages=self.memories['openrouter'],
            )
        except Exception as e:
            print(f"OpenAI API Error: {e}")
        return completion.choices[0].message.content

    def _perplexity_chat(self):
        headers = {
            "accept": "application/json",
            "content-type": "application/json",
            "Authorization": f"Bearer {self.clients['perplexity']['api_key']}"
        }
        payload = {
            "model": "pplx-70b-online",
            "messages": self.memories['perplexity']
        }
        try:
            response = requests.post(self.clients['perplexity']['url'], json=payload, headers=headers)
        except Exception as e:
            print(f"Perplexity API Error: {e}")
        response = json.loads(response.text)
        return response["choices"][0]["message"]["content"]

    def _gemini_chat(self):
        response = self.gemini_chat.send_message(self.memories['gemini'][-1]['content'], stream=True)
        return ''.join(chunk.text for chunk in response)
    
    def test_each_api(self):
        apis = ['openai', 'together', 'openrouter', 'perplexity', 'gemini']
        for api in apis:            
            print(colored(f"Testing {api}...", 'blue'))
            try:
                print(colored("Adding user message...", 'magenta'))
                self.add_message("user", "Hello, how are you?", api)
                print(colored("User message added successfully.", 'green'))
                
                print("Starting chat...")
                response = self.chat_main("Hello, how are you?", api)
                print(colored(f"API successfully responded with: {response}", 'green'))
                
                print(colored("Adding assistant message...", 'yellow'))
                print(colored(F"ALL TESTS WITH {api} PASSED SUCCESSFULLY!", 'green'))
            except Exception as e:
                print(colored(f"An error occurred while testing {api}: {e}", 'red'))
            print("\n")
            print("----------------------------------------")

if __name__ == "__main__":
    # initialize the chatbot and test each model
    chatbot = UnifiedChatBot()
    # chatbot.test_each_api()